#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Package that contains classes for integration with Comet."""

try:
    import comet_ml
except AttributeError:
    # Python 3.6
    comet_ml = None
except ModuleNotFoundError:
    # Python 3.7+
    comet_ml = None

import threading
import time
import os


class CometLogger:
    """Logger class for Comet integration

    `Comet <https://www.comet.ml/>`__ is a meta machine learning platform
    designed to help AI practitioners and teams build reliable machine learning
    models for real-world applications by streamlining the machine learning
    model lifecycle. By leveraging Comet, users can track, compare, explain and
    reproduce their machine learning experiments.

    To log a Spark NLP annotator, it will need an "outputLogPath" parameter, as the
    CometLogger reads the log file generated during the training process.

    For more examples see the `Examples
    <https://github.com/JohnSnowLabs/spark-nlp/blob/master/examples/python/logging/Comet_SparkNLP_Integration.ipynb>`__.

    Parameters
    ----------
    workspace : str, optional
        Name of the workspace in Comet, by default None
    project_name : str, optional
        Name of the project in Comet, by default None
    comet_mode : str, optional
        Mode of logging, by default None. If set to "offline" then offline mode
        will be used, otherwise online.
    experiment_id : str, optional
        Id of the experiment, if it is reused, by default None
    tags : List[str], optional
        List of tags for the experiment, by default None

    Attributes
    ----------
    experiment : comet_ml.Experiment
        Object representing the Comet experiment

    Raises
    ------
    ImportError
        If the package comet-ml is not installed

    Examples
    --------
    Metrics while training an annotator can be logged with for example:

    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from sparknlp.logging.comet import CometLogger
    >>> spark = sparknlp.start()

    To run an online experiment, the logger is defined like so.

    >>> OUTPUT_LOG_PATH = "./run"
    >>> logger = CometLogger()

    Then the experiment can start like so

    >>> document = DocumentAssembler() \\
    ...     .setInputCol("text")\\
    ...     .setOutputCol("document")
    >>> embds = UniversalSentenceEncoder.pretrained() \\
    ...     .setInputCols("document") \\
    ...     .setOutputCol("sentence_embeddings")
    >>> multiClassifier = MultiClassifierDLApproach() \\
    ...     .setInputCols("sentence_embeddings") \\
    ...     .setOutputCol("category") \\
    ...     .setLabelColumn("labels") \\
    ...     .setBatchSize(128) \\
    ...     .setLr(1e-3) \\
    ...     .setThreshold(0.5) \\
    ...     .setShufflePerEpoch(False) \\
    ...     .setEnableOutputLogs(True) \\
    ...     .setOutputLogsPath(OUTPUT_LOG_PATH) \\
    ...     .setMaxEpochs(1)
    >>> logger.monitor(logdir=OUTPUT_LOG_PATH, model=multiClassifier)
    >>> trainDataset = spark.createDataFrame(
    ...     [("Nice.", ["positive"]), ("That's bad.", ["negative"])],
    ...     schema=["text", "labels"],
    ... )
    >>> pipeline = Pipeline(stages=[document, embds, multiClassifier])
    >>> pipeline.fit(trainDataset)
    >>> logger.end()

    If you are using a jupyter notebook, it is possible to display the live web
    interface with

    >>> logger.experiment.display(tab='charts')
    """

    def __init__(
        self,
        workspace=None,
        project_name=None,
        comet_mode=None,
        experiment_id=None,
        tags=None,
        **experiment_kwargs,
    ):
        if comet_ml is None:
            raise ImportError(
                "`comet_ml` is not installed. Please install it with `pip install comet-ml`."
            )

        self.comet_mode = comet_mode
        self.workspace = workspace
        self.project_name = project_name
        self.experiment_id = experiment_id
        self.experiment_kwargs = experiment_kwargs

        self.experiment = self._get_experiment(
            self.comet_mode,
            self.workspace,
            self.project_name,
            self.experiment_id,
            **self.experiment_kwargs,
        )
        self.experiment.log_other("Created from", "SparkNLP")
        if tags is not None:
            self.experiment.add_tags(tags)

        self._watch_file = False
        self._monitor_thread_timeout = 5
        self.thread = None

    def _get_experiment(
        self,
        mode,
        workspace=None,
        project_name=None,
        experiment_id=None,
        **experiment_kwargs,
    ):
        if mode == "offline":
            if experiment_id is not None:
                return comet_ml.ExistingOfflineExperiment(
                    previous_experiment=experiment_id,
                    workspace=workspace,
                    project_name=project_name,
                    **experiment_kwargs,
                )

            return comet_ml.OfflineExperiment(
                workspace=workspace,
                project_name=project_name,
                **experiment_kwargs,
            )

        else:
            if experiment_id is not None:
                return comet_ml.ExistingExperiment(
                    previous_experiment=experiment_id,
                    workspace=workspace,
                    project_name=project_name,
                    **experiment_kwargs,
                )

            return comet_ml.Experiment(
                workspace=workspace,
                project_name=project_name,
                **experiment_kwargs,
            )

    def log_pipeline_parameters(self, pipeline, stages=None):
        """Iterates over the different stages in a pyspark PipelineModel object
        and logs the parameters to Comet.

        Parameters
        ----------
        pipeline : pyspark.ml.PipelineModel
            PipelineModel object
        stages : List[str], optional
            Names of the stages of the pipeline to include, by default None (logs all)

        Examples
        --------
        The pipeline model contains the annotators of Spark NLP, that were
        fitted to a dataframe.

        >>> logger.log_pipeline_parameters(pipeline_model)
        """
        self.experiment.log_other("pipeline_uid", pipeline.uid)
        if stages is None:
            stages = [s.name for s in pipeline.stages]

        for stage in pipeline.stages:
            if stage.name not in stages:
                continue

            params = stage.extractParamMap()
            for param, param_value in params.items():
                self.experiment.log_parameter(f"{stage.name}-{param.name}", param_value)

    def log_visualization(self, html, name="viz.html"):
        """Uploads a NER visualization from Spark NLP Display to comet.

        Parameters
        ----------
        html : str
            HTML of the spark NLP Display visualization
        name : str, optional
            Name for the visualization in comet, by default "viz.html"

        Examples
        --------
        This example has NER chunks (NER extracted by e.g. :class:`.NerDLModel`
        and converted by a :class:`.NerConverter`) extracted in the colum
        "ner_chunk".

        >>> from sparknlp_display import NerVisualizer
        >>> logger = CometLogger()
        >>> for idx, result in enumerate(results.collect()):
        ...     viz = NerVisualizer().display(
        ...         result=result,
        ...         label_col='ner_chunk',
        ...         document_col='document',
        ...         return_html=True
        ...     )
        ...     logger.log_visualization(viz, name=f'viz-{idx}.html')
        """
        self.log_asset_data(html, name)

    def log_metrics(self, metrics, step=None, epoch=None, prefix=None):
        """Submits logs of an evaluation metrics.

        Parameters
        ----------
        metrics : dict
            Dictionary with key value pairs corresponding to the measured metric
            and its value
        step : int, optional
            Used to associate a specific step, by default None
        epoch : int, optional
            Used to associate a specific epoch, by default None
        prefix : str, optional
            Name prefix for this metric, by default None. This can be used to
            identify for example different features by name.

        Examples
        --------
        In this example, sklearn is used to retrieve the metrics.

        >>> from sklearn.preprocessing import MultiLabelBinarizer
        >>> from sklearn.metrics import classification_report
        >>> prediction = model.transform(testDataset)
        >>> preds_df = prediction.select('labels', 'category.result').toPandas()

        >>> mlb = MultiLabelBinarizer()
        >>> y_true = mlb.fit_transform(preds_df['labels'])
        >>> y_pred = mlb.fit_transform(preds_df['result'])
        >>> report = classification_report(y_true, y_pred, output_dict=True)

        Iterate over the report and log the metrics:

        >>> for key, value in report.items():
        ...     logger.log_metrics(value, prefix=key)
        >>> logger.end()

        If you are using Spark NLP in a notebook, then you can display the
        metrics directly with

        >>> logger.experiment.display(tab='metrics')
        """
        self.experiment.log_metrics(metrics, step=step, epoch=epoch, prefix=prefix)

    def log_parameters(self, parameters, step=None):
        """Logs a dictionary (or dictionary-like object) of multiple parameters.

        Parameters
        ----------
        parameters : dict
            Parameters in a key : value form
        step : int, optional
            Used to associate a specific step, by default None, by default None
        """
        self.experiment.log_parameters(parameters, step=step)

    def log_completed_run(self, log_file_path):
        """Submit logs of training metrics after a run has completed.

        Parameters
        ----------
        log_file_path : str
            Path to log file containing training metrics
        """
        with open(log_file_path, "r") as f:
            stats = f.read().splitlines()

        self._parse_log_entry(stats)
        self.experiment.log_other("log_file_path", log_file_path)

    def log_asset(self, asset_path, metadata=None, step=None):
        """Uploads an asset to comet.

        Parameters
        ----------
        asset_path : str
            Path to the asset
        metadata : str, optional
            Some additional data to attach to the the audio asset. Must be a
            JSON-encodable dict, by default None
        step : int, optional
            Used to associate a specific step, by default None, by default None
        """
        self.experiment.log_asset(asset_path, metadata=metadata, step=step)

    def log_asset_data(self, asset, name, overwrite=False, metadata=None, step=None):
        """Uploads the data given to comet (str, binary, or JSON).

        Parameters
        ----------
        asset : str or bytes or dict
            Data to be saved as asset
        name : str
            A custom file name to be displayed
        overwrite : bool, optional
            If True will overwrite all existing assets with the same name, by
            default False
        metadata : dict, optional
            Some additional data to attach to the the asset data.
            Must be a JSON-encodable dict, by default None
        step : int, optional
            Used to associate a specific step, by default None, by default None
        """
        self.experiment.log_asset_data(
            asset, name, overwrite=overwrite, metadata=metadata, step=step
        )

    def monitor(self, logdir, model, interval=10):
        """Monitors the training of the model and submits logs to comet, given
        by an interval.

        To log a Spark NLP annotator, it will need an "outputLogPath" parameter, as the
        CometLogger reads the log file generated during the training process.

        If you are not able to monitor the live training, you can still log the training
        at the end with :meth:`.log_completed_run`.

        Parameters
        ----------
        logdir : str
            Path to the output of the logs
        model : AnnotatorApproach
            Annotator to monitor
        interval : int, optional
            Interval for refreshing, by default 10
        """
        self._watch_file = True
        self.experiment.log_other("model_uid", model.uid)
        self.thread = threading.Thread(
            target=self._monitor_log_file,
            args=(
                os.path.join(logdir, f"{model.uid}.log"),
                interval,
            ),
        )
        self.thread.start()

    def _file_watcher(self, filename, interval):
        """Generator that yields lines from the model log file.

        Parameters
        ----------
        filename : str
            Path to model log file
        interval : int
            Time (seconds) to wait in between checking for file updates

        Yields
        ------
        str
            A single line from the file
        """
        fp = open(filename)

        line = ""
        while self._watch_file:
            partial_line = fp.readline()
            if len(partial_line) != 0:
                line += partial_line
                if line.endswith("\n"):
                    yield line
                    line = ""
            else:
                time.sleep(interval)

        fp.close()

    def _monitor_log_file(self, filename, interval):
        # Wait for file to be created:
        while not os.path.exists(filename) and self._watch_file:
            time.sleep(interval)

        watcher = self._file_watcher(filename, interval)
        for line in watcher:
            lines = line.split("\n")
            self._parse_log_entry(lines)

    def _convert_log_entry_to_dict(self, log_entries):
        output_dict = {}
        for entry in log_entries:
            key, value = entry.strip(" ").split(":")
            output_dict[key] = float(value)

        return output_dict

    def _parse_run_metrics(self, parts):
        epoch_str, ratio = parts[0].split(" ", 1)
        epoch, total = ratio.split("/", 1)

        metrics = parts[2:]
        formatted_metrics = self._convert_log_entry_to_dict(metrics)

        return formatted_metrics, epoch

    def _parse_run_parameters(self, parts):
        parameters = parts[2:]
        formatted_parameters = self._convert_log_entry_to_dict(parameters)
        return formatted_parameters

    def _parse_log_entry(self, lines):
        for line in lines:
            parts = line.split("-")
            if line.startswith("Training started"):
                parameters = self._parse_run_parameters(parts)
                self.log_parameters(parameters)

            elif line.startswith("Epoch"):
                metrics, epoch = self._parse_run_metrics(parts)
                self.log_metrics(metrics, step=int(epoch), epoch=int(epoch))

    def end(self):
        """Ends the experiment and the logger. Submits all outstanding logs to
        comet.
        """
        self._watch_file = False
        self.experiment.end()
        if self.thread:
            self.thread.join(timeout=self._monitor_thread_timeout)



#  Copyright 2017-2024 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sparknlp.common import *

class Gemma3ForMultiModal(AnnotatorModel,
                 HasBatchedAnnotateImage,
                 HasImageFeatureProperties,
                 HasEngine,
                 HasGeneratorProperties):
    """Gemma3ForMultiModal can load Gemma 3 Vision models for visual question answering.
    The model consists of a vision encoder, a text encoder, a text decoder and a model merger.
    The vision encoder will encode the input image, the text encoder will encode the input text,
    the model merger will merge the image and text embeddings, and the text decoder will output the answer.

    Gemma 3 is a family of lightweight, state-of-the-art open models from Google, built from the same 
    research and technology used to create the Gemini models. It features:
    - Large 128K context window
    - Multilingual support in over 140 languages
    - Multimodal capabilities handling both text and image inputs
    - Optimized for deployment on limited resources (laptops, desktops, cloud)

    Pretrained models can be loaded with :meth:`.pretrained` of the companion
    object:

    >>> visualQA = Gemma3ForMultiModal.pretrained() \
    ...     .setInputCols(["image_assembler"]) \
    ...     .setOutputCol("answer")

    The default model is ``"gemma3_4b_it_int4"``, if no name is
    provided.

    For available pretrained models please see the `Models Hub
    <https://sparknlp.org/models?task=Question+Answering>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``IMAGE``              ``DOCUMENT``
    ====================== ======================

    Parameters
    ----------
    batchSize
        Batch size. Large values allows faster processing but requires more
        memory, by default 1
    minOutputLength
        Minimum length of the sequence to be generated, by default 0
    maxOutputLength
        Maximum length of output text, by default 20
    doSample
        Whether or not to use sampling; use greedy decoding otherwise, by default False
    temperature
        The value used to module the next token probabilities, by default 0.6
    topK
        The number of highest probability vocabulary tokens to keep for top-k-filtering, by default -1
    topP
        If set to float < 1, only the most probable tokens with probabilities that add up to ``top_p`` or higher are kept for generation, by default 0.9
    repetitionPenalty
        The parameter for repetition penalty. 1.0 means no penalty, by default 1.0
    noRepeatNgramSize
        If set to int > 0, all ngrams of that size can only occur once, by default 3
    beamSize
        The Number of beams for beam search, by default 1
    maxInputLength
        Maximum length of input text, by default 4096

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> from pyspark.sql.functions import lit
    >>> imageDF = spark.read.format("image").load(images_path)
    >>> testDF = imageDF.withColumn("text", lit("<bos><start_of_turn>user\\nYou are a helpful assistant.\\n\\n<start_of_image>Describe this image in detail.<end_of_turn>\\n<start_of_turn>model\\n"))
    >>> imageAssembler = ImageAssembler() \\
    ...     .setInputCol("image") \\
    ...     .setOutputCol("image_assembler")
    >>> visualQA = Gemma3ForMultiModal.pretrained() \\
    ...     .setInputCols("image_assembler") \\
    ...     .setOutputCol("answer")
    >>> pipeline = Pipeline().setStages([
    ...     imageAssembler,
    ...     visualQA
    ... ])
    >>> result = pipeline.fit(testDF).transform(testDF)
    >>> result.select("image_assembler.origin", "answer.result").show(truncate=False)
    """

    name = "Gemma3ForMultiModal"

    inputAnnotatorTypes = [AnnotatorType.IMAGE]

    outputAnnotatorType = AnnotatorType.DOCUMENT

    configProtoBytes = Param(Params._dummy(),
                             "configProtoBytes",
                             "ConfigProto from tensorflow, serialized into byte array. Get with "
                             "config_proto.SerializeToString()",
                             TypeConverters.toListInt)

    minOutputLength = Param(Params._dummy(), "minOutputLength", "Minimum length of the sequence to be generated",
                            typeConverter=TypeConverters.toInt)

    maxOutputLength = Param(Params._dummy(), "maxOutputLength", "Maximum length of output text",
                            typeConverter=TypeConverters.toInt)

    doSample = Param(Params._dummy(), "doSample", "Whether or not to use sampling; use greedy decoding otherwise",
                     typeConverter=TypeConverters.toBoolean)

    temperature = Param(Params._dummy(), "temperature", "The value used to module the next token probabilities",
                        typeConverter=TypeConverters.toFloat)

    topK = Param(Params._dummy(), "topK",
                 "The number of highest probability vocabulary tokens to keep for top-k-filtering",
                 typeConverter=TypeConverters.toInt)

    topP = Param(Params._dummy(), "topP",
                 "If set to float < 1, only the most probable tokens with probabilities that add up to ``top_p`` or higher are kept for generation",
                 typeConverter=TypeConverters.toFloat)

    repetitionPenalty = Param(Params._dummy(), "repetitionPenalty",
                              "The parameter for repetition penalty. 1.0 means no penalty. See `this paper <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details",
                              typeConverter=TypeConverters.toFloat)

    noRepeatNgramSize = Param(Params._dummy(), "noRepeatNgramSize",
                              "If set to int > 0, all ngrams of that size can only occur once",
                              typeConverter=TypeConverters.toInt)

    ignoreTokenIds = Param(Params._dummy(), "ignoreTokenIds",
                           "A list of token ids which are ignored in the decoder's output",
                           typeConverter=TypeConverters.toListInt)
    beamSize = Param(Params._dummy(), "beamSize",
                     "The Number of beams for beam search.",
                     typeConverter=TypeConverters.toInt)

    maxInputLength = Param(Params._dummy(), "maxInputLength", "Maximum length of input text",
                            typeConverter=TypeConverters.toInt)

    def setMaxSentenceSize(self, value):
        """Sets Maximum sentence length that the annotator will process, by
        default 50.

        Parameters
        ----------
        value : int
            Maximum sentence length that the annotator will process
        """
        return self._set(maxSentenceLength=value)

    def setIgnoreTokenIds(self, value):
        """A list of token ids which are ignored in the decoder's output.

        Parameters
        ----------
        value : List[int]
            The words to be filtered out
        """
        return self._set(ignoreTokenIds=value)

    def setConfigProtoBytes(self, b):
        """Sets configProto from tensorflow, serialized into byte array.

        Parameters
        ----------
        b : List[int]
            ConfigProto from tensorflow, serialized into byte array
        """
        return self._set(configProtoBytes=b)

    def setMinOutputLength(self, value):
        """Sets minimum length of the sequence to be generated.

        Parameters
        ----------
        value : int
            Minimum length of the sequence to be generated
        """
        return self._set(minOutputLength=value)

    def setMaxOutputLength(self, value):
        """Sets maximum length of output text.

        Parameters
        ----------
        value : int
            Maximum length of output text
        """
        return self._set(maxOutputLength=value)

    def setDoSample(self, value):
        """Sets whether or not to use sampling, use greedy decoding otherwise.

        Parameters
        ----------
        value : bool
            Whether or not to use sampling; use greedy decoding otherwise
        """
        return self._set(doSample=value)

    def setTemperature(self, value):
        """Sets the value used to module the next token probabilities.

        Parameters
        ----------
        value : float
            The value used to module the next token probabilities
        """
        return self._set(temperature=value)

    def setTopK(self, value):
        """Sets the number of highest probability vocabulary tokens to keep for
        top-k-filtering.

        Parameters
        ----------
        value : int
            Number of highest probability vocabulary tokens to keep
        """
        return self._set(topK=value)

    def setTopP(self, value):
        """Sets the top cumulative probability for vocabulary tokens.

        If set to float < 1, only the most probable tokens with probabilities
        that add up to ``topP`` or higher are kept for generation.

        Parameters
        ----------
        value : float
            Cumulative probability for vocabulary tokens
        """
        return self._set(topP=value)

    def setRepetitionPenalty(self, value):
        """Sets the parameter for repetition penalty. 1.0 means no penalty.

        Parameters
        ----------
        value : float
            The repetition penalty

        References
        ----------
        See `Ctrl: A Conditional Transformer Language Model For Controllable
        Generation <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details.
        """
        return self._set(repetitionPenalty=value)

    def setNoRepeatNgramSize(self, value):
        """Sets size of n-grams that can only occur once.

        If set to int > 0, all ngrams of that size can only occur once.

        Parameters
        ----------
        value : int
            N-gram size can only occur once
        """
        return self._set(noRepeatNgramSize=value)

    def setBeamSize(self, value):
        """Sets the number of beam size for beam search, by default `4`.

        Parameters
        ----------
        value : int
            Number of beam size for beam search
        """
        return self._set(beamSize=value)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.cv.Gemma3ForMultiModal",
                 java_model=None):
        super(Gemma3ForMultiModal, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            batchSize=1,
            minOutputLength=0,
            maxOutputLength=20,
            doSample=False,
            temperature=0.6,
            topK=-1,
            topP=0.9,
            repetitionPenalty=1.0,
            noRepeatNgramSize=3,
            ignoreTokenIds=[],
            beamSize=1,
            maxInputLength=4096,
        )

    @staticmethod
    def loadSavedModel(folder, spark_session, use_openvino=False):
        """Loads a locally saved model.

        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession

        Returns
        -------
        Gemma3ForMultiModal
            The restored model
        """
        from sparknlp.internal import _Gemma3ForMultiModalLoader
        jModel = _Gemma3ForMultiModalLoader(folder, spark_session._jsparkSession, use_openvino)._java_obj
        return Gemma3ForMultiModal(java_model=jModel)

    @staticmethod
    def pretrained(name="gemma3_4b_it_int4", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.

        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default "gemma3_4b_it_int4"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.

        Returns
        -------
        Gemma3ForMultiModal
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(Gemma3ForMultiModal, name, lang, remote_loc) 


#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the Finisher."""

from pyspark import keyword_only
from pyspark.ml.param import TypeConverters, Params, Param
from sparknlp.internal import AnnotatorTransformer


class Finisher(AnnotatorTransformer):
    """Converts annotation results into a format that easier to use.

    It is useful to extract the results from Spark NLP Pipelines. The Finisher
    outputs annotation(s) values into ``String``.

    For more extended examples on document pre-processing see the
    `Examples <https://github.com/JohnSnowLabs/spark-nlp/blob/master/examples/python/annotation/text/english/model-downloader/Create%20custom%20pipeline%20-%20NerDL.ipynb
>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``ANY``                ``NONE``
    ====================== ======================

    Parameters
    ----------
    inputCols
        Input annotations
    outputCols
        Output finished annotation cols
    valueSplitSymbol
        Character separating values, by default #
    annotationSplitSymbol
        Character separating annotations, by default @
    cleanAnnotations
        Whether to remove annotation columns, by default True
    includeMetadata
        Whether to include annotation metadata, by default False
    outputAsArray
        Finisher generates an Array with the results instead of string, by
        default True
    parseEmbeddingsVectors
        Whether to include embeddings vectors in the process, by default False

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from sparknlp.pretrained import PretrainedPipeline
    >>> data = spark.createDataFrame([[1, "New York and New Jersey aren't that far apart actually."]]).toDF("id", "text")

    Define pretrained pipeline that extracts Named Entities amongst other things
    and apply the `Finisher` on it.

    >>> pipeline = PretrainedPipeline("explain_document_dl")
    >>> finisher = Finisher().setInputCols("entities").setOutputCols("output")
    >>> explainResult = pipeline.transform(data)

    Show results.

    >>> explainResult.selectExpr("explode(entities)").show(truncate=False)
    +------------------------------------------------------------------------------------------------------------------------------------------------------+
    |entities                                                                                                                                              |
    +------------------------------------------------------------------------------------------------------------------------------------------------------+
    |[[chunk, 0, 7, New York, [entity -> LOC, sentence -> 0, chunk -> 0], []], [chunk, 13, 22, New Jersey, [entity -> LOC, sentence -> 0, chunk -> 1], []]]|
    +------------------------------------------------------------------------------------------------------------------------------------------------------+
    >>> result = finisher.transform(explainResult)
    >>> result.select("output").show(truncate=False)
    +----------------------+
    |output                |
    +----------------------+
    |[New York, New Jersey]|
    +----------------------+

    See Also
    --------
    Finisher : for finishing Strings
    """

    inputCols = Param(Params._dummy(), "inputCols", "input annotations", typeConverter=TypeConverters.toListString)
    outputCols = Param(Params._dummy(), "outputCols", "output finished annotation cols", typeConverter=TypeConverters.toListString)
    valueSplitSymbol = Param(Params._dummy(), "valueSplitSymbol", "character separating annotations", typeConverter=TypeConverters.toString)
    annotationSplitSymbol = Param(Params._dummy(), "annotationSplitSymbol", "character separating annotations", typeConverter=TypeConverters.toString)
    cleanAnnotations = Param(Params._dummy(), "cleanAnnotations", "whether to remove annotation columns", typeConverter=TypeConverters.toBoolean)
    includeMetadata = Param(Params._dummy(), "includeMetadata", "annotation metadata format", typeConverter=TypeConverters.toBoolean)
    outputAsArray = Param(Params._dummy(), "outputAsArray", "finisher generates an Array with the results instead of string", typeConverter=TypeConverters.toBoolean)
    parseEmbeddingsVectors = Param(Params._dummy(), "parseEmbeddingsVectors", "whether to include embeddings vectors in the process", typeConverter=TypeConverters.toBoolean)
    name = "Finisher"

    @keyword_only
    def __init__(self):
        super(Finisher, self).__init__(classname="com.johnsnowlabs.nlp.Finisher")
        self._setDefault(
            cleanAnnotations=True,
            includeMetadata=False,
            outputAsArray=True,
            parseEmbeddingsVectors=False,
            valueSplitSymbol="#",
            annotationSplitSymbol="@",
            outputCols=[]
        )

    @keyword_only
    def setParams(self):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setInputCols(self, *value):
        """Sets column names of input annotations.

        Parameters
        ----------
        *value : List[str]
            Input columns for the annotator
        """
        if len(value) == 1 and type(value[0]) == list:
            return self._set(inputCols=value[0])
        else:
            return self._set(inputCols=list(value))

    def setOutputCols(self, *value):
        """Sets column names of finished output annotations.

        Parameters
        ----------
        *value : List[str]
            List of output columns
        """
        if len(value) == 1 and type(value[0]) == list:
            return self._set(outputCols=value[0])
        else:
            return self._set(outputCols=list(value))

    def setValueSplitSymbol(self, value):
        """Sets character separating values, by default #.

        Parameters
        ----------
        value : str
            Character to separate annotations
        """
        return self._set(valueSplitSymbol=value)

    def setAnnotationSplitSymbol(self, value):
        """Sets character separating annotations, by default @.

        Parameters
        ----------
        value : str
            ...
        """
        return self._set(annotationSplitSymbol=value)

    def setCleanAnnotations(self, value):
        """Sets whether to remove annotation columns, by default True.

        Parameters
        ----------
        value : bool
            Whether to remove annotation columns
        """
        return self._set(cleanAnnotations=value)

    def setIncludeMetadata(self, value):
        """Sets whether to include annotation metadata.

        Parameters
        ----------
        value : bool
            Whether to include annotation metadata
        """
        return self._set(includeMetadata=value)

    def setOutputAsArray(self, value):
        """Sets whether to generate an array with the results instead of a
        string.

        Parameters
        ----------
        value : bool
            Whether to generate an array with the results instead of a string
        """
        return self._set(outputAsArray=value)

    def setParseEmbeddingsVectors(self, value):
        """Sets whether to include embeddings vectors in the process.

        Parameters
        ----------
        value : bool
            Whether to include embeddings vectors in the process
        """
        return self._set(parseEmbeddingsVectors=value)

    def getInputCols(self):
        """Gets input columns name of annotations."""
        return self.getOrDefault(self.inputCols)

    def getOutputCols(self):
        """Gets output columns name of annotations."""
        if len(self.getOrDefault(self.outputCols)) == 0:
            return ["finished_" + input_col for input_col in self.getInputCols()]
        else:
            return self.getOrDefault(self.outputCols)



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the ImageAssembler."""

from pyspark import keyword_only
from pyspark.ml.param import TypeConverters, Params, Param
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import regexp_replace, col

from sparknlp.common import AnnotatorType
from sparknlp.internal import AnnotatorTransformer


class ImageAssembler(AnnotatorTransformer):
    """Prepares images read by Spark into a format that is processable by Spark NLP.
    This component is needed to process images.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``NONE``               ``IMAGE``
    ====================== ======================

    Parameters
    ----------
    inputCol
        Input column name
    outputCol
        Output column name

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from pyspark.ml import Pipeline
    >>> data = spark.read.format("image").load("./tmp/images/").toDF("image")
    >>> imageAssembler = ImageAssembler().setInputCol("image").setOutputCol("image_assembler")
    >>> result = imageAssembler.transform(data)
    >>> result.select("image_assembler").show()
    >>> result.select("image_assembler").printSchema()
    root
     |-- image_assembler: array (nullable = true)
     |    |-- element: struct (containsNull = true)
     |    |    |-- annotatorType: string (nullable = true)
     |    |    |-- origin: string (nullable = true)
     |    |    |-- height: integer (nullable = true)
     |    |    |-- width: integer (nullable = true)
     |    |    |-- nChannels: integer (nullable = true)
     |    |    |-- mode: integer (nullable = true)
     |    |    |-- result: binary (nullable = true)
     |    |    |-- metadata: map (nullable = true)
     |    |    |    |-- key: string
     |    |    |    |-- value: string (valueContainsNull = true)
    """

    outputAnnotatorType = AnnotatorType.IMAGE

    inputCol = Param(Params._dummy(), "inputCol", "input column name", typeConverter=TypeConverters.toString)
    textCol = Param(Params._dummy(), "textCol", "text column name", typeConverter=TypeConverters.toString)
    outputCol = Param(Params._dummy(), "outputCol", "output column name", typeConverter=TypeConverters.toString)
    name = 'ImageAssembler'

    @keyword_only
    def __init__(self):
        super(ImageAssembler, self).__init__(classname="com.johnsnowlabs.nlp.ImageAssembler")
        self._setDefault(outputCol="image_assembler", inputCol='image')

    @keyword_only
    def setParams(self):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setInputCol(self, value):
        """Sets input column name.

        Parameters
        ----------
        value : str
            Name of the input column that has image format loaded via spark.read.format("image").load(PATH)
        """
        return self._set(inputCol=value)

    def setOutputCol(self, value):
        """Sets output column name.

        Parameters
        ----------
        value : str
            Name of the Output Column
        """
        return self._set(outputCol=value)

    def getOutputCol(self):
        """Gets output column name of annotations."""
        return self.getOrDefault(self.outputCol)

    def setTextCol(self, value):
        """Sets an optional text column name.

        Parameters
        ----------
        value : str
            Name of an optional input text column
        """
        return self._set(inputCol=value)

    @classmethod
    def loadImagesAsBytes(cls, spark: SparkSession, path: str):
        """
        Loads images from a given path and returns them as raw bytes, instead of the default
        OpenCV-compatible format. Supported image types include JPEG, PNG, GIF, and BMP.

        Multimodal inference with llama.cpp requires raw bytes as input.

        Parameters
        ----------
        spark : SparkSession
            The active SparkSession.
        path : str
            The path to the images. Supported image types are JPEG, PNG, GIF, and BMP.

        Returns
        -------
        DataFrame
            A DataFrame containing the images as raw bytes along with their metadata.
        """

        # Replace the path separator in the `origin` field and `path` column, so that they match
        def replace_path(column_name: str):
            return regexp_replace(col(column_name), ":///", ":/")

        # Load the images as metadata with the default Spark image format
        data = (
            spark.read.format("image")
            .option("dropInvalid", True)
            .load(path)
            .withColumn(
                "image", col("image").withField("origin", replace_path("image.origin"))
            )
        )

        # Load the images as raw binary files
        image_bytes = (
            spark.read.format("binaryFile")
            .option("pathGlobFilter", "*.{jpeg,jpg,png,gif,bmp,JPEG,JPG,PNG,GIF,BMP}")
            .option("dropInvalid", True)
            .load(path)
            .withColumn("path", replace_path("path"))
        )

        # Join the two datasets on the file path
        df_joined = data.join(
            image_bytes, data["image.origin"] == image_bytes["path"], "inner"
        )

        # Replace the `data` field of the `image` column with raw bytes
        df_image_replaced = df_joined.withColumn(
            "image", df_joined["image"].withField("data", df_joined["content"])
        )

        return df_image_replaced



from sparknlp.common import *

class InternVLForMultiModal(AnnotatorModel,
                          HasBatchedAnnotateImage,
                          HasImageFeatureProperties,
                          HasEngine,
                          HasGeneratorProperties):
    """
    InternVLForMultiModal can load InternVL Vision models for visual question answering.
    The model consists of a vision encoder, a text encoder, a text decoder and a model merger.
    The vision encoder will encode the input image, the text encoder will encode the input text,
    the model merger will merge the image and text embeddings, and the text decoder will output the answer.

    InternVL 2.5 is an advanced multimodal large language model (MLLM) series that builds upon InternVL 2.0,
    maintaining its core model architecture while introducing significant enhancements in training and testing
    strategies as well as data quality. Key features include:
    - Large context window support
    - Multilingual support
    - Multimodal capabilities handling both text and image inputs
    - Optimized for deployment with int4 quantization

    Pretrained models can be loaded with :meth:`.pretrained` of the companion object:
    >>> visualQA = InternVLForMultiModal.pretrained() \\
    ...     .setInputCols("image_assembler") \\
    ...     .setOutputCol("answer")

    The default model is `"internvl2_5_1b_int4"`, if no name is provided.
    For available pretrained models, refer to the `Models Hub
    <https://sparknlp.org/models?task=Question+Answering>`__.

    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``IMAGE``              ``DOCUMENT``
    ====================== ======================

    Parameters
    ----------
    batchSize : int, optional
        Batch size. Larger values allow faster processing but require more memory,
        by default 1.
    maxSentenceLength : int, optional
        Maximum sentence length to process, by default 4096.

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> from pyspark.sql.functions import lit
    >>> image_df = spark.read.format("image").load(path=images_path)
    >>> test_df = image_df.withColumn(
    ...     "text",
    ...     lit("<|im_start|><image>\\nDescribe this image in detail.<|im_end|><|im_start|>assistant\\n")
    ... )
    >>> imageAssembler = ImageAssembler() \\
    ...     .setInputCol("image") \\
    ...     .setOutputCol("image_assembler")
    >>> visualQA = InternVLForMultiModal.pretrained() \\
    ...     .setInputCols("image_assembler") \\
    ...     .setOutputCol("answer")
    >>> pipeline = Pipeline().setStages([
    ...     imageAssembler,
    ...     visualQA
    ... ])

    >>> result = pipeline.fit(test_df).transform(test_df)
    >>> result.select("image_assembler.origin", "answer.result").show(truncate=False)
    """

    name = "InternVLForMultiModal"

    inputAnnotatorTypes = [AnnotatorType.IMAGE]

    outputAnnotatorType = AnnotatorType.DOCUMENT

    minOutputLength = Param(Params._dummy(), "minOutputLength", "Minimum length of the sequence to be generated",
                            typeConverter=TypeConverters.toInt)

    maxOutputLength = Param(Params._dummy(), "maxOutputLength", "Maximum length of output text",
                            typeConverter=TypeConverters.toInt)

    doSample = Param(Params._dummy(), "doSample", "Whether or not to use sampling; use greedy decoding otherwise",
                     typeConverter=TypeConverters.toBoolean)

    temperature = Param(Params._dummy(), "temperature", "The value used to module the next token probabilities",
                        typeConverter=TypeConverters.toFloat)

    topK = Param(Params._dummy(), "topK",
                 "The number of highest probability vocabulary tokens to keep for top-k-filtering",
                 typeConverter=TypeConverters.toInt)

    topP = Param(Params._dummy(), "topP",
                 "If set to float < 1, only the most probable tokens with probabilities that add up to ``top_p`` or higher are kept for generation",
                 typeConverter=TypeConverters.toFloat)

    repetitionPenalty = Param(Params._dummy(), "repetitionPenalty",
                              "The parameter for repetition penalty. 1.0 means no penalty. See `this paper <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details",
                              typeConverter=TypeConverters.toFloat)

    noRepeatNgramSize = Param(Params._dummy(), "noRepeatNgramSize",
                              "If set to int > 0, all ngrams of that size can only occur once",
                              typeConverter=TypeConverters.toInt)

    ignoreTokenIds = Param(Params._dummy(), "ignoreTokenIds",
                           "A list of token ids which are ignored in the decoder's output",
                           typeConverter=TypeConverters.toListInt)

    beamSize = Param(Params._dummy(), "beamSize",
                     "The Number of beams for beam search.",
                     typeConverter=TypeConverters.toInt)

    def setMaxSentenceSize(self, value):
        """Sets Maximum sentence length that the annotator will process, by
        default 4096.
        Parameters
        ----------
        value : int
            Maximum sentence length that the annotator will process
        """
        return self._set(maxSentenceLength=value)

    def setIgnoreTokenIds(self, value):
        """A list of token ids which are ignored in the decoder's output.
        Parameters
        ----------
        value : List[int]
            The words to be filtered out
        """
        return self._set(ignoreTokenIds=value)

    def setMinOutputLength(self, value):
        """Sets minimum length of the sequence to be generated.
        Parameters
        ----------
        value : int
            Minimum length of the sequence to be generated
        """
        return self._set(minOutputLength=value)

    def setMaxOutputLength(self, value):
        """Sets maximum length of output text.
        Parameters
        ----------
        value : int
            Maximum length of output text
        """
        return self._set(maxOutputLength=value)

    def setDoSample(self, value):
        """Sets whether or not to use sampling, use greedy decoding otherwise.
        Parameters
        ----------
        value : bool
            Whether or not to use sampling; use greedy decoding otherwise
        """
        return self._set(doSample=value)

    def setTemperature(self, value):
        """Sets the value used to module the next token probabilities.
        Parameters
        ----------
        value : float
            The value used to module the next token probabilities
        """
        return self._set(temperature=value)

    def setTopK(self, value):
        """Sets the number of highest probability vocabulary tokens to keep for
        top-k-filtering.
        Parameters
        ----------
        value : int
            Number of highest probability vocabulary tokens to keep
        """
        return self._set(topK=value)

    def setTopP(self, value):
        """Sets the top cumulative probability for vocabulary tokens.
        If set to float < 1, only the most probable tokens with probabilities
        that add up to ``topP`` or higher are kept for generation.
        Parameters
        ----------
        value : float
            Cumulative probability for vocabulary tokens
        """
        return self._set(topP=value)

    def setRepetitionPenalty(self, value):
        """Sets the parameter for repetition penalty. 1.0 means no penalty.
        Parameters
        ----------
        value : float
            The repetition penalty
        References
        ----------
        See `Ctrl: A Conditional Transformer Language Model For Controllable
        Generation <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details.
        """
        return self._set(repetitionPenalty=value)

    def setNoRepeatNgramSize(self, value):
        """Sets size of n-grams that can only occur once.
        If set to int > 0, all ngrams of that size can only occur once.
        Parameters
        ----------
        value : int
            N-gram size can only occur once
        """
        return self._set(noRepeatNgramSize=value)

    def setBeamSize(self, value):
        """Sets the number of beam size for beam search, by default `1`.
        Parameters
        ----------
        value : int
            Number of beam size for beam search
        """
        return self._set(beamSize=value)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.cv.InternVLForMultiModal",
                 java_model=None):
        super(InternVLForMultiModal, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            batchSize=1,
            minOutputLength=0,
            maxOutputLength=20,
            doSample=False,
            temperature=0.6,
            topK=-1,
            topP=0.9,
            repetitionPenalty=1.0,
            noRepeatNgramSize=3,
            ignoreTokenIds=[],
            beamSize=1
        )

    @staticmethod
    def loadSavedModel(folder, spark_session, use_openvino=False):
        """Loads a locally saved model.
        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession
        Returns
        -------
        InternVLForMultiModal
            The restored model
        """
        from sparknlp.internal import _InternVLForMultiModalLoader
        jModel = _InternVLForMultiModalLoader(folder, spark_session._jsparkSession, use_openvino)._java_obj
        return InternVLForMultiModal(java_model=jModel)

    @staticmethod
    def pretrained(name="internvl2_5_1b_int4", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.
        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default
            "internvl2_5_1b_int4"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.
        Returns
        -------
        InternVLForMultiModal
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(InternVLForMultiModal, name, lang, remote_loc) 



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for MiniLMEmbeddings."""

from sparknlp.common import *


class MiniLMEmbeddings(AnnotatorModel,
                           HasEmbeddingsProperties,
                           HasCaseSensitiveProperties,
                           HasStorageRef,
                           HasBatchedAnnotate,
                           HasMaxSentenceLengthLimit):
    """Sentence embeddings using MiniLM.

    MiniLM, a lightweight and efficient sentence embedding model that can generate text embeddings for various NLP tasks (e.g., classification, retrieval, clustering, text evaluation, etc.)
    Note that this annotator is only supported for Spark Versions 3.4 and up.

    Pretrained models can be loaded with :meth:`.pretrained` of the companion
    object:

    >>> embeddings = MiniLMEmbeddings.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("minilm_embeddings")


    The default model is ``"minilm_l6_v2"``, if no name is provided.

    For available pretrained models please see the
    `Models Hub <https://sparknlp.org/models?q=MiniLM>`__.


    ====================== ======================
    Input Annotation types Output Annotation type
    ====================== ======================
    ``DOCUMENT``            ``SENTENCE_EMBEDDINGS``
    ====================== ======================

    Parameters
    ----------
    batchSize
        Size of every batch , by default 8
    dimension
        Number of embedding dimensions, by default 384
    caseSensitive
        Whether to ignore case in tokens for embeddings matching, by default False
    maxSentenceLength
        Max sentence length to process, by default 512
    configProtoBytes
        ConfigProto from tensorflow, serialized into byte array.

    References
    ----------
    `MiniLM: Deep Self-Attention Distillation for Task-Agnostic Compression of Pre-Trained Transformers <https://arxiv.org/abs/2002.10957>`__

    `MiniLM Github Repository <https://github.com/microsoft/unilm/tree/master/minilm>`__

    **Paper abstract**

    *We present a simple and effective approach to compress large pre-trained Transformer models
    by distilling the self-attention module of the last Transformer layer. The compressed model
    (called MiniLM) can be trained with task-agnostic distillation and then fine-tuned on various
    downstream tasks. We evaluate MiniLM on the GLUE benchmark and show that it achieves comparable
    results with BERT-base while being 4.3x smaller and 5.5x faster. We also show that MiniLM can
    be further compressed to 22x smaller and 12x faster than BERT-base while maintaining comparable
    performance.*

    Examples
    --------
    >>> import sparknlp
    >>> from sparknlp.base import *
    >>> from sparknlp.annotator import *
    >>> from pyspark.ml import Pipeline
    >>> documentAssembler = DocumentAssembler() \\
    ...     .setInputCol("text") \\
    ...     .setOutputCol("document")
    >>> embeddings = MiniLMEmbeddings.pretrained() \\
    ...     .setInputCols(["document"]) \\
    ...     .setOutputCol("minilm_embeddings")
    >>> embeddingsFinisher = EmbeddingsFinisher() \\
    ...     .setInputCols(["minilm_embeddings"]) \\
    ...     .setOutputCols("finished_embeddings") \\
    ...     .setOutputAsVector(True)
    >>> pipeline = Pipeline().setStages([
    ...     documentAssembler,
    ...     embeddings,
    ...     embeddingsFinisher
    ... ])
    >>> data = spark.createDataFrame([["This is a sample sentence for embedding generation.",
    ... "Another example sentence to demonstrate MiniLM embeddings.",
    ... ]]).toDF("text")
    >>> result = pipeline.fit(data).transform(data)
    >>> result.selectExpr("explode(finished_embeddings) as result").show(5, 80)
    +--------------------------------------------------------------------------------+
    |                                                                          result|
    +--------------------------------------------------------------------------------+
    |[[0.1234567, -0.2345678, 0.3456789, -0.4567890, 0.5678901, -0.6789012...|
    |[[0.2345678, -0.3456789, 0.4567890, -0.5678901, 0.6789012, -0.7890123...|
    +--------------------------------------------------------------------------------+
    """

    name = "MiniLMEmbeddings"

    inputAnnotatorTypes = [AnnotatorType.DOCUMENT]

    outputAnnotatorType = AnnotatorType.SENTENCE_EMBEDDINGS
    configProtoBytes = Param(Params._dummy(),
                             "configProtoBytes",
                             "ConfigProto from tensorflow, serialized into byte array. Get with config_proto.SerializeToString()",
                             TypeConverters.toListInt)


    def setConfigProtoBytes(self, b):
        """Sets configProto from tensorflow, serialized into byte array.

        Parameters
        ----------
        b : List[int]
            ConfigProto from tensorflow, serialized into byte array
        """
        return self._set(configProtoBytes=b)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.embeddings.MiniLMEmbeddings", java_model=None):
        super(MiniLMEmbeddings, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            dimension=384,
            batchSize=8,
            maxSentenceLength=512,
            caseSensitive=False,
        )

    @staticmethod
    def loadSavedModel(folder, spark_session, use_openvino=False):
        """Loads a locally saved model.

        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession
        use_openvino : bool
            Use OpenVINO backend

        Returns
        -------
        MiniLMEmbeddings
            The restored model
        """
        from sparknlp.internal import _MiniLMLoader
        jModel = _MiniLMLoader(folder, spark_session._jsparkSession, use_openvino)._java_obj
        return MiniLMEmbeddings(java_model=jModel)

    @staticmethod
    def pretrained(name="minilm_l6_v2", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.

        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default "minilm_l6_v2"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.

        Returns
        -------
        MiniLMEmbeddings
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(MiniLMEmbeddings, name, lang, remote_loc)



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module of annotators for text classification."""
from sparknlp.annotator.classifier_dl.albert_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.albert_for_token_classification import *
from sparknlp.annotator.classifier_dl.albert_for_question_answering import *
from sparknlp.annotator.classifier_dl.bert_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.bert_for_token_classification import *
from sparknlp.annotator.classifier_dl.bert_for_question_answering import *
from sparknlp.annotator.classifier_dl.classifier_dl import *
from sparknlp.annotator.classifier_dl.deberta_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.deberta_for_token_classification import *
from sparknlp.annotator.classifier_dl.deberta_for_question_answering import *
from sparknlp.annotator.classifier_dl.distil_bert_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.distil_bert_for_token_classification import *
from sparknlp.annotator.classifier_dl.distil_bert_for_question_answering import *
from sparknlp.annotator.classifier_dl.longformer_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.longformer_for_token_classification import *
from sparknlp.annotator.classifier_dl.longformer_for_question_answering import *
from sparknlp.annotator.classifier_dl.multi_classifier_dl import *
from sparknlp.annotator.classifier_dl.roberta_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.roberta_for_token_classification import *
from sparknlp.annotator.classifier_dl.roberta_for_question_answering import *
from sparknlp.annotator.classifier_dl.sentiment_dl import *
from sparknlp.annotator.classifier_dl.xlm_roberta_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.xlm_roberta_for_token_classification import *
from sparknlp.annotator.classifier_dl.xlm_roberta_for_question_answering import *
from sparknlp.annotator.classifier_dl.xlnet_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.xlnet_for_token_classification import *
from sparknlp.annotator.classifier_dl.camembert_for_token_classification import *
from sparknlp.annotator.classifier_dl.tapas_for_question_answering import *
from sparknlp.annotator.classifier_dl.camembert_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.camembert_for_question_answering import *
from sparknlp.annotator.classifier_dl.bert_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.distil_bert_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.roberta_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.xlm_roberta_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.bart_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.deberta_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.mpnet_for_sequence_classification import *
from sparknlp.annotator.classifier_dl.mpnet_for_question_answering import *
from sparknlp.annotator.classifier_dl.mpnet_for_token_classification import *
from sparknlp.annotator.classifier_dl.albert_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.camembert_for_zero_shot_classification import *
from sparknlp.annotator.classifier_dl.bert_for_multiple_choice import *
from sparknlp.annotator.classifier_dl.xlm_roberta_for_multiple_choice import *
from sparknlp.annotator.classifier_dl.roberta_for_multiple_choice import *
from sparknlp.annotator.classifier_dl.distilbert_for_multiple_choice import *
from sparknlp.annotator.classifier_dl.albert_for_multiple_choice import *



#  Copyright 2017-2024 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sparknlp.internal import ExtendedJavaWrapper


class SparkNLPReader(ExtendedJavaWrapper):
    """Instantiates class to read documents in various formats.

    Parameters
    ----------
    params : spark
        Spark session
    params : dict, optional
        Parameter with custom configuration

    Notes
    -----
    This class can read HTML, email, PDF, MS Word, Excel, PowerPoint, and text files.

    Examples
    --------
    >>> from sparknlp.reader import SparkNLPReader
    >>> reader = SparkNLPReader(spark)

    Reading HTML

    >>> html_df = reader.html("https://www.wikipedia.org")
    >>> # Or with shorthand
    >>> import sparknlp
    >>> html_df = sparknlp.read().html("https://www.wikipedia.org")

    Reading PDF

    >>> pdf_df = reader.pdf("home/user/pdfs-directory")
    >>> # Or with shorthand
    >>> pdf_df = sparknlp.read().pdf("home/user/pdfs-directory")

    Reading Email
    
    >>> email_df = reader.email("home/user/emails-directory")
    >>> # Or with shorthand
    >>> email_df = sparknlp.read().email("home/user/emails-directory")
    """

    def __init__(self, spark, params=None, headers=None):
        if params is None:
            params = {}
        super(SparkNLPReader, self).__init__("com.johnsnowlabs.reader.SparkNLPReader", params, headers)
        self.spark = spark

    def html(self, htmlPath):
        """Reads HTML files or URLs and returns a Spark DataFrame.

        Parameters
        ----------
        htmlPath : str or list of str
            Path(s) to HTML file(s) or a list of URLs.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing the parsed HTML content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> html_df = SparkNLPReader().html("https://www.wikipedia.org")

        You can also use SparkNLP to simplify the process:

        >>> import sparknlp
        >>> html_df = sparknlp.read().html("https://www.wikipedia.org")
        >>> html_df.show(truncate=False)

        +--------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |url                 |html                                                                                                                                                                                                                                                                                                                            |
        +--------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |https://example.com/|[{Title, Example Domain, {pageNumber -> 1}}, {NarrativeText, 0, This domain is for use in illustrative examples in documents. You may use this domain in literature without prior coordination or asking for permission., {pageNumber -> 1}}, {NarrativeText, 0, More information... More information..., {pageNumber -> 1}}]   |
        +--------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        >>> html_df.printSchema()
        root
         |-- url: string (nullable = true)
         |-- html: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)
        """
        if not isinstance(htmlPath, (str, list)) or (isinstance(htmlPath, list) and not all(isinstance(item, str) for item in htmlPath)):
            raise TypeError("htmlPath must be a string or a list of strings")
        jdf = self._java_obj.html(htmlPath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def email(self, filePath):
        """Reads email files and returns a Spark DataFrame.

        Parameters
        ----------
        filePath : str
            Path to an email file or a directory containing emails.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed email data.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> email_df = SparkNLPReader(spark).email("home/user/emails-directory")

        You can also use SparkNLP to simplify the process:

        >>> import sparknlp
        >>> email_df = sparknlp.read().email("home/user/emails-directory")
        >>> email_df.show()
        +---------------------------------------------------+
        |email                                              |
        +---------------------------------------------------+
        |[{Title, Email Text Attachments, {sent_to -> Danilo|
        +---------------------------------------------------+
        >>> email_df.printSchema()
        root
         |-- path: string (nullable = true)
         |-- content: array (nullable = true)
         |-- email: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)

        """
        if not isinstance(filePath, str):
            raise TypeError("filePath must be a string")
        jdf = self._java_obj.email(filePath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def doc(self, docPath):
        """Reads word document files and returns a Spark DataFrame.

        Parameters
        ----------
        docPath : str
            Path to a word document file.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed document content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> doc_df = SparkNLPReader().doc(spark, "home/user/word-directory")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> doc_df = sparknlp.read().doc("home/user/word-directory")
        >>> doc_df.show()
        +-------------------------------------------------+
        |doc                                              |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                |
        +-------------------------------------------------+
        |[{Table, Header Col 1, {}}, {Table, Header Col 2,|
        +-------------------------------------------------+

        >>> doc_df.printSchema()
        root
         |-- path: string (nullable = true)
         |-- content: array (nullable = true)
         |-- doc: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)

        """
        if not isinstance(docPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.doc(docPath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def epub(self, epubPath):
        """Reads EPUB files and returns a Spark DataFrame.

        Parameters
        ----------
        epubPath : str
            Path to an EPUB file or a directory containing EPUB files.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed EPUB content.
        """
        if not isinstance(epubPath, str):
            raise TypeError("epubPath must be a string")
        jdf = self._java_obj.epub(epubPath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def pdf(self, pdfPath):
        if not isinstance(pdfPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.pdf(pdfPath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def xls(self, docPath):
        """Reads excel document files and returns a Spark DataFrame.

        Parameters
        ----------
        docPath : str
            Path to an excel document file.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed document content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> xlsDf = SparkNLPReader().xls(spark, "home/user/excel-directory")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> xlsDf = sparknlp.read().xls("home/user/excel-directory")
        >>> xlsDf.show()
        +--------------------------------------------+
        |xls                                         |
        +--------------------------------------------+
        |[{Title, Financial performance, {SheetNam}}]|
        +--------------------------------------------+

        >>> xlsDf.printSchema()
        root
         |-- path: string (nullable = true)
         |-- content: binary (nullable = true)
         |-- xls: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)
       """
        if not isinstance(docPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.xls(docPath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def ppt(self, docPath):
        """
        Reads power point document files and returns a Spark DataFrame.

        Parameters
        ----------
        docPath : str
            Path to an power point document file.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed document content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> pptDf = SparkNLPReader().ppt(spark, "home/user/powerpoint-directory")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> pptDf = sparknlp.read().ppt("home/user/powerpoint-directory")
        >>> pptDf.show(truncate=False)
        +-------------------------------------+
        |ppt                                  |
        +-------------------------------------+
        |[{Title, Adding a Bullet Slide, {}},]|
        +-------------------------------------+
        """
        if not isinstance(docPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.ppt(docPath)
        dataframe = self.getDataFrame(self.spark, jdf)
        return dataframe

    def txt(self, docPath):
        """Reads TXT files and returns a Spark DataFrame.

        Parameters
        ----------
        docPath : str
            Path to a TXT file.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed document content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> txtDf = SparkNLPReader().txt(spark, "home/user/txt/files")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> txtDf = sparknlp.read().txt("home/user/txt/files")
        >>> txtDf.show(truncate=False)
        +-----------------------------------------------+
        |txt                                            |
        +-----------------------------------------------+
        |[{Title, BIG DATA ANALYTICS, {paragraph -> 0}}]|
        +-----------------------------------------------+
        """
        if not isinstance(docPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.txt(docPath)
        return self.getDataFrame(self.spark, jdf)

    def xml(self, docPath):
        """Reads XML files and returns a Spark DataFrame.

        Parameters
        ----------
        docPath : str
            Path to an XML file or a directory containing XML files.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed XML content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> xml_df = SparkNLPReader(spark).xml("home/user/xml-directory")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> xml_df = sparknlp.read().xml("home/user/xml-directory")
        >>> xml_df.show(truncate=False)
        +-----------------------------------------------------------+
        |xml                                                       |
        +-----------------------------------------------------------+
        |[{Title, John Smith, {elementId -> ..., tag -> title}}]   |
        +-----------------------------------------------------------+

        >>> xml_df.printSchema()
        root
         |-- path: string (nullable = true)
         |-- xml: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)
        """
        if not isinstance(docPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.xml(docPath)
        return self.getDataFrame(self.spark, jdf)


    def md(self, filePath):
        """Reads Markdown files and returns a Spark DataFrame.

        Parameters
        ----------
        filePath : str
            Path to a Markdown file or a directory containing Markdown files.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed Markdown content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> md_df = SparkNLPReader(spark).md("home/user/markdown-directory")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> md_df = sparknlp.read().md("home/user/markdown-directory")
        >>> md_df.show(truncate=False)
        +-----------------------------------------------------------+
        |md                                                         |
        +-----------------------------------------------------------+
        |[{Title, Sample Markdown Document, {elementId -> ..., tag -> title}}]|
        +-----------------------------------------------------------+

        >>> md_df.printSchema()
        root
         |-- path: string (nullable = true)
         |-- md: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)
        """
        if not isinstance(filePath, str):
            raise TypeError("filePath must be a string")
        jdf = self._java_obj.md(filePath)
        return self.getDataFrame(self.spark, jdf)

    def csv(self, csvPath):
        """Reads CSV files and returns a Spark DataFrame.

        Parameters
        ----------
        docPath : str
            Path to an CSV file or a directory containing CSV files.

        Returns
        -------
        pyspark.sql.DataFrame
            A DataFrame containing parsed CSV content.

        Examples
        --------
        >>> from sparknlp.reader import SparkNLPReader
        >>> csv_df = SparkNLPReader(spark).csv("home/user/csv-directory")

        You can use SparkNLP for one line of code

        >>> import sparknlp
        >>> csv_df = sparknlp.read().csv("home/user/csv-directory")
        >>> csv_df.show(truncate=False)
        +-----------------------------------------------------------------------------------------------------------------------------------------+
        |csv                                                                                                                                      |
        +-----------------------------------------------------------------------------------------------------------------------------------------+
        |[{NarrativeText, Alice 100 Bob 95, {}}, {Table, <table><tr><td>Alice</td><td>100</td></tr><tr><td>Bob</td><td>95</td></tr></table>, {}}] |
        +-----------------------------------------------------------------------------------------------------------------------------------------+

        >>> csv_df.printSchema()
        root
         |-- path: string (nullable = true)
         |-- csv: array (nullable = true)
         |    |-- element: struct (containsNull = true)
         |    |    |-- elementType: string (nullable = true)
         |    |    |-- content: string (nullable = true)
         |    |    |-- metadata: map (nullable = true)
         |    |    |    |-- key: string
         |    |    |    |-- value: string (valueContainsNull = true)
        """
        if not isinstance(csvPath, str):
            raise TypeError("docPath must be a string")
        jdf = self._java_obj.csv(csvPath)
        return self.getDataFrame(self.spark, jdf)



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for MxbaiEmbeddings."""

from sparknlp.common import *


class MxbaiEmbeddings(AnnotatorModel,
					HasEmbeddingsProperties,
					HasCaseSensitiveProperties,
					HasStorageRef,
					HasBatchedAnnotate,
					HasMaxSentenceLengthLimit):
	"""Sentence embeddings using Mxbai Embeddings.


	Pretrained models can be loaded with :meth:`.pretrained` of the companion
	object:

	>>> embeddings = MxbaiEmbeddings.pretrained() \\
	...     .setInputCols(["document"]) \\
	...     .setOutputCol("Mxbai_embeddings")


	The default model is ``"mxbai_large_v1"``, if no name is provided.

	For available pretrained models please see the
	`Models Hub <https://sparknlp.org/models?q=Mxbai>`__.


	====================== ======================
	Input Annotation types Output Annotation type
	====================== ======================
	``DOCUMENT``            ``SENTENCE_EMBEDDINGS``
	====================== ======================

	Parameters
	----------
	batchSize
		Size of every batch , by default 8
	dimension
		Number of embedding dimensions, by default 768
	caseSensitive
		Whether to ignore case in tokens for embeddings matching, by default False
	maxSentenceLength
		Max sentence length to process, by default 512
	configProtoBytes
		ConfigProto from tensorflow, serialized into byte array.



	Examples
	--------
	>>> import sparknlp
	>>> from sparknlp.base import *
	>>> from sparknlp.annotator import *
	>>> from pyspark.ml import Pipeline
	>>> documentAssembler = DocumentAssembler() \\
	...     .setInputCol("text") \\
	...     .setOutputCol("document")
	>>> embeddings = MxbaiEmbeddings.pretrained() \\
	...     .setInputCols(["document"]) \\
	...     .setOutputCol("embeddings")
	>>> embeddingsFinisher = EmbeddingsFinisher() \\
	...     .setInputCols("embeddings") \\
	...     .setOutputCols("finished_embeddings") \\
	...     .setOutputAsVector(True)
	>>> pipeline = Pipeline().setStages([
	...     documentAssembler,
	...     embeddings,
	...     embeddingsFinisher
	... ])
	>>> data = spark.createDataFrame([["hello world", "hello moon"]]).toDF("text")
	>>> result = pipeline.fit(data).transform(data)
	>>> result.selectExpr("explode(finished_embeddings) as result").show(5, 80)
	+--------------------------------------------------------------------------------+
	|                                                                          result|
	+--------------------------------------------------------------------------------+
	|[0.50387806, 0.5861606, 0.35129607, -0.76046336, -0.32446072, -0.117674336, 0...|
	|[0.6660665, 0.961762, 0.24854276, -0.1018044, -0.6569202, 0.027635604, 0.1915...|
	+--------------------------------------------------------------------------------+
	"""

	name = "MxbaiEmbeddings"

	inputAnnotatorTypes = [AnnotatorType.DOCUMENT]

	outputAnnotatorType = AnnotatorType.SENTENCE_EMBEDDINGS
	poolingStrategy = Param(Params._dummy(),
							"poolingStrategy",
							"Pooling strategy to use for sentence embeddings",
							TypeConverters.toString)

	def setPoolingStrategy(self, value):
		"""Pooling strategy to use for sentence embeddings.

		Available pooling strategies for sentence embeddings are:
		  - `"cls"`: leading `[CLS]` token
		  - `"cls_avg"`: leading `[CLS]` token + mean of all other tokens
		  - `"last"`: embeddings of the last token in the sequence
		  - `"avg"`: mean of all tokens
		  - `"max"`: max of all embedding features of the entire token sequence
		  - `"int"`: An integer number, which represents the index of the token to use as the
			embedding

		Parameters
		----------
		value : str
			Pooling strategy to use for sentence embeddings
		"""

		valid_strategies = {"cls", "cls_avg", "last", "avg", "max"}
		if value in valid_strategies or value.isdigit():
			return self._set(poolingStrategy=value)
		else:
			raise ValueError(f"Invalid pooling strategy: {value}. "
							 f"Valid strategies are: {', '.join(self.valid_strategies)} or an integer.")

	@keyword_only
	def __init__(self, classname="com.johnsnowlabs.nlp.embeddings.MxbaiEmbeddings", java_model=None):
		super(MxbaiEmbeddings, self).__init__(
			classname=classname,
			java_model=java_model
		)
		self._setDefault(
			dimension=1024,
			batchSize=8,
			maxSentenceLength=512,
			caseSensitive=False,
			poolingStrategy="cls"
		)

	@staticmethod
	def loadSavedModel(folder, spark_session):
		"""Loads a locally saved model.

		Parameters
		----------
		folder : str
			Folder of the saved model
		spark_session : pyspark.sql.SparkSession
			The current SparkSession

		Returns
		-------
		MxbaiEmbeddings
			The restored model
		"""
		from sparknlp.internal import _MxbaiEmbeddingsLoader
		jModel = _MxbaiEmbeddingsLoader(folder, spark_session._jsparkSession)._java_obj
		return MxbaiEmbeddings(java_model=jModel)

	@staticmethod
	def pretrained(name="mxbai_large_v1", lang="en", remote_loc=None):
		"""Downloads and loads a pretrained model.

		Parameters
		----------
		name : str, optional
			Name of the pretrained model, by default "mxbai_large_v1"
		lang : str, optional
			Language of the pretrained model, by default "en"
		remote_loc : str, optional
			Optional remote address of the resource, by default None. Will use
			Spark NLPs repositories otherwise.

		Returns
		-------
		MxbaiEmbeddings
			The restored model
		"""
		from sparknlp.pretrained import ResourceDownloader
		return ResourceDownloader.downloadModel(MxbaiEmbeddings, name, lang, remote_loc)



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the CoHereTransformer."""

from sparknlp.common import *


class CoHereTransformer(AnnotatorModel, HasBatchedAnnotate, HasEngine):
    """Cohere: Command-R Transformer
    
        C4AI Command-R is a research release of a 35 billion parameter highly performant generative model.
        Command-R is a large language model with open weights optimized for a variety of use cases including reasoning,
        summarization, and question answering. Command-R has the capability for multilingual generation evaluated
        in 10 languages and highly performant RAG capabilities.

        Pretrained models can be loaded with :meth:`.pretrained` of the companion
        object:
    
        >>> CoHere = CoHereTransformer.pretrained() \\
        ...     .setInputCols(["document"]) \\
        ...     .setOutputCol("generation")
    
    
        The default model is ``"c4ai_command_r_v01_int4"``, if no name is provided. For available
        pretrained models please see the `Models Hub
        <https://sparknlp.org/models?q=CoHere>`__.
    
        ====================== ======================
        Input Annotation types Output Annotation type
        ====================== ======================
        ``DOCUMENT``           ``DOCUMENT``
        ====================== ======================
    
        Parameters
        ----------
        configProtoBytes
            ConfigProto from tensorflow, serialized into byte array.
        minOutputLength
            Minimum length of the sequence to be generated, by default 0
        maxOutputLength
            Maximum length of output text, by default 60
        doSample
            Whether or not to use sampling; use greedy decoding otherwise, by default False
        temperature
            The value used to modulate the next token probabilities, by default 1.0
        topK
            The number of highest probability vocabulary tokens to keep for
            top-k-filtering, by default 40
        topP
            Top cumulative probability for vocabulary tokens, by default 1.0
    
            If set to float < 1, only the most probable tokens with probabilities
            that add up to ``topP`` or higher are kept for generation.
        repetitionPenalty
            The parameter for repetition penalty, 1.0 means no penalty. , by default
            1.0
        noRepeatNgramSize
            If set to int > 0, all ngrams of that size can only occur once, by
            default 0
        ignoreTokenIds
            A list of token ids which are ignored in the decoder's output, by
            default []
    
        Notes
        -----
        This is a very computationally expensive module, especially on larger
        sequences. The use of an accelerator such as GPU is recommended.
    
        References
        ----------
        - `Cohere <https://cohere.for.ai/>`__

    
        Examples
        --------
        >>> import sparknlp
        >>> from sparknlp.base import *
        >>> from sparknlp.annotator import *
        >>> from pyspark.ml import Pipeline
        >>> documentAssembler = DocumentAssembler() \\
        ...     .setInputCol("text") \\
        ...     .setOutputCol("documents")
        >>> CoHere = CoHereTransformer.pretrained("c4ai_command_r_v01_int4","en") \\
        ...     .setInputCols(["documents"]) \\
        ...     .setMaxOutputLength(60) \\
        ...     .setOutputCol("generation")
        >>> pipeline = Pipeline().setStages([documentAssembler, CoHere])
        >>> data = spark.createDataFrame([
        ...     (
        ...         1,
        ...         "<BOS_TOKEN><|START_OF_TURN_TOKEN|><|USER_TOKEN|>Hello, how are you?<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>"
        ...         )
        ... ]).toDF("id", "text")
        >>> result = pipeline.fit(data).transform(data)
        >>> result.select("generation.result").show(truncate=False)
        +------------------------------------------------+
        |result                                          |
        +------------------------------------------------+
        |[Hello! I'm doing well, thank you for asking! I'm excited to help you with whatever questions you have today. How can I assist you?]|
        +------------------------------------------------+
    """

    name = "CoHereTransformer"

    inputAnnotatorTypes = [AnnotatorType.DOCUMENT]

    outputAnnotatorType = AnnotatorType.DOCUMENT

    configProtoBytes = Param(Params._dummy(),
                             "configProtoBytes",
                             "ConfigProto from tensorflow, serialized into byte array. Get with config_proto.SerializeToString()",
                             TypeConverters.toListInt)

    minOutputLength = Param(Params._dummy(), "minOutputLength", "Minimum length of the sequence to be generated",
                            typeConverter=TypeConverters.toInt)

    maxOutputLength = Param(Params._dummy(), "maxOutputLength", "Maximum length of output text",
                            typeConverter=TypeConverters.toInt)

    doSample = Param(Params._dummy(), "doSample", "Whether or not to use sampling; use greedy decoding otherwise",
                     typeConverter=TypeConverters.toBoolean)

    temperature = Param(Params._dummy(), "temperature", "The value used to module the next token probabilities",
                        typeConverter=TypeConverters.toFloat)

    topK = Param(Params._dummy(), "topK",
                 "The number of highest probability vocabulary tokens to keep for top-k-filtering",
                 typeConverter=TypeConverters.toInt)

    topP = Param(Params._dummy(), "topP",
                 "If set to float < 1, only the most probable tokens with probabilities that add up to ``top_p`` or higher are kept for generation",
                 typeConverter=TypeConverters.toFloat)

    repetitionPenalty = Param(Params._dummy(), "repetitionPenalty",
                              "The parameter for repetition penalty. 1.0 means no penalty. See `this paper <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details",
                              typeConverter=TypeConverters.toFloat)

    noRepeatNgramSize = Param(Params._dummy(), "noRepeatNgramSize",
                              "If set to int > 0, all ngrams of that size can only occur once",
                              typeConverter=TypeConverters.toInt)

    ignoreTokenIds = Param(Params._dummy(), "ignoreTokenIds",
                           "A list of token ids which are ignored in the decoder's output",
                           typeConverter=TypeConverters.toListInt)

    beamSize = Param(Params._dummy(), "beamSize",
                     "The number of beams to use for beam search",
                     typeConverter=TypeConverters.toInt)

    stopTokenIds = Param(Params._dummy(), "stopTokenIds",
                         "A list of token ids which are considered as stop tokens in the decoder's output",
                         typeConverter=TypeConverters.toListInt)

    def setIgnoreTokenIds(self, value):
        """A list of token ids which are ignored in the decoder's output.

        Parameters
        ----------
        value : List[int]
            The words to be filtered out
        """
        return self._set(ignoreTokenIds=value)

    def setConfigProtoBytes(self, b):
        """Sets configProto from tensorflow, serialized into byte array.

        Parameters
        ----------
        b : List[int]
            ConfigProto from tensorflow, serialized into byte array
        """
        return self._set(configProtoBytes=b)

    def setMinOutputLength(self, value):
        """Sets minimum length of the sequence to be generated.

        Parameters
        ----------
        value : int
            Minimum length of the sequence to be generated
        """
        return self._set(minOutputLength=value)

    def setMaxOutputLength(self, value):
        """Sets maximum length of output text.

        Parameters
        ----------
        value : int
            Maximum length of output text
        """
        return self._set(maxOutputLength=value)

    def setDoSample(self, value):
        """Sets whether or not to use sampling, use greedy decoding otherwise.

        Parameters
        ----------
        value : bool
            Whether or not to use sampling; use greedy decoding otherwise
        """
        return self._set(doSample=value)

    def setTemperature(self, value):
        """Sets the value used to module the next token probabilities.

        Parameters
        ----------
        value : float
            The value used to module the next token probabilities
        """
        return self._set(temperature=value)

    def setTopK(self, value):
        """Sets the number of highest probability vocabulary tokens to keep for
        top-k-filtering.

        Parameters
        ----------
        value : int
            Number of highest probability vocabulary tokens to keep
        """
        return self._set(topK=value)

    def setTopP(self, value):
        """Sets the top cumulative probability for vocabulary tokens.

        If set to float < 1, only the most probable tokens with probabilities
        that add up to ``topP`` or higher are kept for generation.

        Parameters
        ----------
        value : float
            Cumulative probability for vocabulary tokens
        """
        return self._set(topP=value)

    def setRepetitionPenalty(self, value):
        """Sets the parameter for repetition penalty. 1.0 means no penalty.

        Parameters
        ----------
        value : float
            The repetition penalty

        References
        ----------
        See `Ctrl: A Conditional Transformer Language Model For Controllable
        Generation <https://arxiv.org/pdf/1909.05858.pdf>`__ for more details.
        """
        return self._set(repetitionPenalty=value)

    def setNoRepeatNgramSize(self, value):
        """Sets size of n-grams that can only occur once.

        If set to int > 0, all ngrams of that size can only occur once.

        Parameters
        ----------
        value : int
            N-gram size can only occur once
        """
        return self._set(noRepeatNgramSize=value)

    def setBeamSize(self, value):
        """Sets the number of beams to use for beam search.

        Parameters
        ----------
        value : int
            The number of beams to use for beam search
        """
        return self._set(beamSize=value)

    def setStopTokenIds(self, value):
        """Sets a list of token ids which are considered as stop tokens in the decoder's output.

        Parameters
        ----------
        value : List[int]
            The words to be considered as stop tokens
        """
        return self._set(stopTokenIds=value)

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.seq2seq.CoHereTransformer", java_model=None):
        super(CoHereTransformer, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            minOutputLength=0,
            maxOutputLength=20,
            doSample=False,
            temperature=0.6,
            topK=-1,
            topP=0.9,
            repetitionPenalty=1.0,
            noRepeatNgramSize=3,
            ignoreTokenIds=[],
            batchSize=1,
            beamSize=1,
            stopTokenIds=[128001, ]
        )

    @staticmethod
    def loadSavedModel(folder, spark_session, use_openvino=False):
        """Loads a locally saved model.

        Parameters
        ----------
        folder : str
            Folder of the saved model
        spark_session : pyspark.sql.SparkSession
            The current SparkSession

        Returns
        -------
        CoHereTransformer
            The restored model
        """
        from sparknlp.internal import _CoHereLoader
        jModel = _CoHereLoader(folder, spark_session._jsparkSession, use_openvino)._java_obj
        return CoHereTransformer(java_model=jModel)

    @staticmethod
    def pretrained(name="c4ai_command_r_v01_int4", lang="en", remote_loc=None):
        """Downloads and loads a pretrained model.

        Parameters
        ----------
        name : str, optional
            Name of the pretrained model, by default "c4ai_command_r_v01_int4"
        lang : str, optional
            Language of the pretrained model, by default "en"
        remote_loc : str, optional
            Optional remote address of the resource, by default None. Will use
            Spark NLPs repositories otherwise.

        Returns
        -------
        CoHereTransformer
            The restored model
        """
        from sparknlp.pretrained import ResourceDownloader
        return ResourceDownloader.downloadModel(CoHereTransformer, name, lang, remote_loc)



#  Copyright 2017-2022 John Snow Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Contains classes for the LightPipeline."""

import sparknlp.internal as _internal
from sparknlp.annotation import Annotation
from sparknlp.annotation_audio import AnnotationAudio
from sparknlp.annotation_image import AnnotationImage
from sparknlp.common import AnnotatorApproach, AnnotatorModel
from sparknlp.internal import AnnotatorTransformer


class LightPipeline:
    """Creates a LightPipeline from a Spark PipelineModel.

    LightPipeline is a Spark NLP specific Pipeline class equivalent to Spark
    ML Pipeline. The difference is that it’s execution does not hold to
    Spark principles, instead it computes everything locally (but in
    parallel) in order to achieve fast results when dealing with small
    amounts of data. This means, we do not input a Spark Dataframe, but a
    string or an Array of strings instead, to be annotated. To create Light
    Pipelines, you need to input an already trained (fit) Spark ML Pipeline.

    It’s :meth:`.transform` has now an alternative :meth:`.annotate`, which
    directly outputs the results.

    Parameters
    ----------
    pipelineModel : :class:`pyspark.ml.PipelineModel`
        The PipelineModel containing Spark NLP Annotators
    parse_embeddings : bool, optional
        Whether to parse embeddings, by default False

    Notes
    -----
    Use :meth:`.fullAnnotate` to also output the result as
    :class:`.Annotation`, with metadata.

    Examples
    --------
    >>> from sparknlp.base import LightPipeline
    >>> light = LightPipeline(pipeline.fit(data))
    >>> light.annotate("We are very happy about Spark NLP")
    {
        'document': ['We are very happy about Spark NLP'],
        'lemmas': ['We', 'be', 'very', 'happy', 'about', 'Spark', 'NLP'],
        'pos': ['PRP', 'VBP', 'RB', 'JJ', 'IN', 'NNP', 'NNP'],
        'sentence': ['We are very happy about Spark NLP'],
        'spell': ['We', 'are', 'very', 'happy', 'about', 'Spark', 'NLP'],
        'stems': ['we', 'ar', 'veri', 'happi', 'about', 'spark', 'nlp'],
        'token': ['We', 'are', 'very', 'happy', 'about', 'Spark', 'NLP']
    }
    """

    def __init__(self, pipelineModel, parse_embeddings=False, output_cols=None):
        """
        Parameters
        ----------
        pipelineModel : PipelineModel
            The fitted Spark NLP pipeline model.
        parse_embeddings : bool, optional
            Whether to parse embeddings.
        output_cols : list[str], optional
            List of output columns to return in results (optional).
        """
        if output_cols is None:
            output_cols = []

        self.pipeline_model = pipelineModel
        self.parse_embeddings = parse_embeddings
        self.output_cols = output_cols

        self._lightPipeline = _internal._LightPipeline(pipelineModel, parse_embeddings, output_cols).apply()

    def _validateStagesInputCols(self, stages):
        annotator_types = self._getAnnotatorTypes(stages)
        for stage in stages:
            if isinstance(stage, AnnotatorApproach) or isinstance(stage, AnnotatorModel):
                input_cols = stage.getInputCols()
                if type(input_cols) == str:
                    input_cols = [input_cols]
                input_annotator_types = stage.inputAnnotatorTypes + stage.optionalInputAnnotatorTypes
                for input_col in input_cols:
                    annotator_type = annotator_types.get(input_col)
                    if annotator_type is None or annotator_type not in input_annotator_types:
                        raise TypeError(f"Wrong or missing inputCols annotators in {stage.uid}"
                                        f" Make sure such annotator exist in your pipeline,"
                                        f" with the right output names and that they have following annotator types:"
                                        f" {input_annotator_types}")

    def _skipPipelineValidation(self, stages):
        exceptional_pipeline = [stage for stage in stages if self._skipStageValidation(stage)]
        if len(exceptional_pipeline) >= 1:
            return True
        else:
            return False

    def _skipStageValidation(self, stage):
        return hasattr(stage, 'skipLPInputColsValidation') and stage.skipLPInputColsValidation

    def _getAnnotatorTypes(self, stages):
        annotator_types = {}
        for stage in stages:
            if hasattr(stage, 'getOutputCols'):
                output_cols = stage.getOutputCols()
                for output_col in output_cols:
                    annotator_types[output_col] = stage.outputAnnotatorType
            elif isinstance(stage, AnnotatorApproach) or isinstance(stage, AnnotatorModel) or\
                    isinstance(stage, AnnotatorTransformer):
                if stage.outputAnnotatorType is not None:
                    annotator_types[stage.getOutputCol()] = stage.outputAnnotatorType
        return annotator_types

    def _annotationFromJava(self, java_annotations):
        annotations = []
        for annotation in java_annotations:

            index = annotation.toString().index("(")
            annotation_type = annotation.toString()[:index]

            if annotation_type == "AnnotationImage":
                result = self.__get_result(annotation)
                annotations.append(
                    AnnotationImage(annotation.annotatorType(),
                                    annotation.origin(),
                                    annotation.height(),
                                    annotation.width(),
                                    annotation.nChannels(),
                                    annotation.mode(),
                                    result,
                                    annotation.metadata())
                )
            elif annotation_type == "AnnotationAudio":
                result = self.__get_result(annotation)
                annotations.append(
                    AnnotationAudio(annotation.annotatorType(),
                                    result,
                                    annotation.metadata())
                )
            else:
                if self.parse_embeddings:
                    embeddings = list(annotation.embeddings())
                else:
                    embeddings = []
                annotations.append(
                    Annotation(annotation.annotatorType(),
                               annotation.begin(),
                               annotation.end(),
                               annotation.result(),
                               annotation.metadata(),
                               embeddings)
                )
        return annotations

    @staticmethod
    def __get_result(annotation):
        try:
            result = list(annotation.result())
        except TypeError:
            result = []

        return result

    @staticmethod
    def __isMetadataInput(value):
        if isinstance(value, dict):
            return True

        return isinstance(value, list) and len(value) > 0 and all(
            isinstance(item, dict) for item in value
        )

    def __extractMetadataFromArgs(self, args, metadata):
        if metadata is not None:
            return args, metadata

        if len(args) == 2 and self.__isMetadataInput(args[1]):
            return (args[0],), args[1]

        return args, metadata

    def __normalizeMetadataForTextInput(self, target, metadata):
        if metadata is None or isinstance(target, str):
            return metadata

        if isinstance(metadata, dict):
            return self.__columnarToRowMetadata(metadata, len(target))

        return metadata

    @staticmethod
    def __columnarToRowMetadata(metadata, target_size):
        if not metadata:
            return [{} for _ in range(target_size)]

        for key, values in metadata.items():
            if not isinstance(values, list):
                raise TypeError(
                    "When passing multiple texts with dict metadata, each metadata "
                    "value must be a list."
                )
            if len(values) != target_size:
                raise TypeError(
                    "When passing multiple texts with dict metadata, each metadata "
                    "list must have the same length as target."
                )

        return [
            {key: [values[idx]] for key, values in metadata.items()}
            for idx in range(target_size)
        ]

    def fullAnnotate(self, *args, **kwargs):
        """
        Annotate and return full Annotation objects.

        Supports both:
          - fullAnnotate(text: str)
          - fullAnnotate(texts: list[str])
          - fullAnnotate(ids: list[int], texts: list[str])
          - fullAnnotate(text: str, metadata: dict[str, list[str]])
          - fullAnnotate(texts: list[str], metadata: list[dict[str, list[str]]])
          - fullAnnotate(texts: list[str], metadata: dict[str, list[str]])


        Examples
        --------
        >>> from sparknlp.pretrained import PretrainedPipeline
        >>> explain_document_pipeline = PretrainedPipeline("explain_document_dl")
        >>> result = explain_document_pipeline.fullAnnotate('U.N. official Ekeus heads for Baghdad.')
        >>> result[0].keys()
        dict_keys(['entities', 'stem', 'checked', 'lemma', 'document', 'pos', 'token', 'ner', 'embeddings', 'sentence'])
        >>> result[0]["ner"]
        [Annotation(named_entity, 0, 2, B-ORG, {'word': 'U.N'}),
        Annotation(named_entity, 3, 3, O, {'word': '.'}),
        Annotation(named_entity, 5, 12, O, {'word': 'official'}),
        Annotation(named_entity, 14, 18, B-PER, {'word': 'Ekeus'}),
        Annotation(named_entity, 20, 24, O, {'word': 'heads'}),
        Annotation(named_entity, 26, 28, O, {'word': 'for'}),
        Annotation(named_entity, 30, 36, B-LOC, {'word': 'Baghdad'}),
        Annotation(named_entity, 37, 37, O, {'word': '.'})]
        """
        metadata = kwargs.pop("metadata", None)

        if "target" in kwargs:
            args = (kwargs["target"],) + args
        if "optional_target" in kwargs:
            args = args + (kwargs["optional_target"],)
        args, metadata = self.__extractMetadataFromArgs(args, metadata)

        stages = self.pipeline_model.stages
        if not self._skipPipelineValidation(stages):
            self._validateStagesInputCols(stages)

        input_type = self.__detectInputType(args)

        if input_type == "ids_texts":
            ids, texts = args
            results = self._lightPipeline.fullAnnotateWithIdsJava(ids, texts)
            return [self.__buildStages(r) for r in results]

        if input_type == "qa":
            question, context = args
            if metadata is not None:
                raise TypeError("metadata is only supported for text inputs.")
            return self.__fullAnnotateQuestionAnswering(question, context)

        if input_type == "text":
            target = args[0]
            metadata = self.__normalizeMetadataForTextInput(target, metadata)
            return self.__fullAnnotateText(target, metadata)

        if input_type == "audio":
            audios = args[0]
            if metadata is not None:
                raise TypeError("metadata is only supported for text inputs.")
            return self.__fullAnnotateAudio(audios)

        if input_type == "image":
            images = args[0]
            if metadata is not None:
                raise TypeError("metadata is only supported for text inputs.")
            return self.fullAnnotateImage(images)

        raise TypeError(
            "Unsupported input for fullAnnotate(). Expected: "
            "(text: str | list[str]), "
            "(ids: list[int], texts: list[str]), "
            "(question: str, context: str), "
            "(audio: list[float] | list[list[float]]), or "
            "(image_path: str | list[str])."
        )

    @staticmethod
    def __isTextInput(target):
        if type(target) is str:
            return True
        elif type(target) is list and type(target[0]) is str:
            return True
        else:
            return False

    @staticmethod
    def __isAudioInput(target):
        if type(target) is list and type(target[0]) is float:
            return True
        elif type(target) is list and type(target[0]) is list and type(target[0][0]) is float:
            return True
        else:
            return False

    def __fullAnnotateText(self, target, metadata=None):

        if self.__isPath(target):
            if metadata is not None:
                raise TypeError("metadata is only supported for text inputs.")
            result = self.fullAnnotateImage(target)
            return result
        else:
            result = []
            if type(target) is str:
                if metadata is None:
                    target = [target]
                    for annotations_result in self._lightPipeline.fullAnnotateJava(target):
                        result.append(self.__buildStages(annotations_result))
                else:
                    annotations_result = self._lightPipeline.fullAnnotateWithMetaJava(
                        target, metadata
                    )
                    result.append(self.__buildStages(annotations_result))
            else:
                if metadata is None:
                    for annotations_result in self._lightPipeline.fullAnnotateJava(target):
                        result.append(self.__buildStages(annotations_result))
                else:
                    if not isinstance(metadata, list):
                        raise TypeError(
                            "metadata must be a list of dicts or a dict of lists when "
                            "passing multiple texts."
                        )
                    if len(metadata) != len(target):
                        raise TypeError("metadata and target must have the same length.")
                    for annotations_result in self._lightPipeline.fullAnnotateWithMetaJava(
                        target, metadata
                    ):
                        result.append(self.__buildStages(annotations_result))
            return result

    def __isPath(self, target):
        if type(target) is list:
            target = target[0]

        if target.find("/") < 0:
            return False
        else:
            is_valid_file = _internal._ResourceHelper_validFile(target).apply()
            return is_valid_file

    def __fullAnnotateAudio(self, audios):
        result = []
        if type(audios[0]) is float:
            annotations_dict = self._lightPipeline.fullAnnotateSingleAudioJava(audios)
            result.append(self.__buildStages(annotations_dict))
        else:
            full_annotations = self._lightPipeline.fullAnnotateAudiosJava(audios)
            for annotations_dict in full_annotations:
                result.append(self.__buildStages(annotations_dict))

        return result

    def __fullAnnotateQuestionAnswering(self, question, context):
        result = []
        if type(question) is str and type(context) is str:
            annotations_dict = self._lightPipeline.fullAnnotateJava(question, context)
            result.append(self.__buildStages(annotations_dict))
        else:
            full_annotations = self._lightPipeline.fullAnnotateJava(question, context)
            for annotations_dict in full_annotations:
                result.append(self.__buildStages(annotations_dict))

        return result

    def fullAnnotateImage(self, path_to_image, text=None):
        """Annotates the data provided into `Annotation` type results.

        The data should be either a list or a str.

        Parameters
        ----------
        path_to_image : list or str
            Source path of image, list of paths to images

        text: list or str, optional
           Optional list or str of texts. If None, defaults to empty list if path_to_image is a list, or empty string if path_to_image is a string.

        Returns
        -------
        List[AnnotationImage]
            The result of the annotation
        """
        if not isinstance(path_to_image, (str, list)):
            raise TypeError("argument for path_to_image must be 'str' or 'list[str]'")

        if text is None:
            text = "" if isinstance(path_to_image, str) else []

        if type(path_to_image) != type(text):
            raise ValueError("`path_to_image` and `text` must be of the same type")

        stages = self.pipeline_model.stages
        if not self._skipPipelineValidation(stages):
            self._validateStagesInputCols(stages)

        if isinstance(path_to_image, str):
            path_to_image = [path_to_image]
            text = [text]

        result = []

        for image_result in self._lightPipeline.fullAnnotateImageJava(path_to_image, text):
            result.append(self.__buildStages(image_result))

        return result


    def __buildStages(self, annotations_result):
        stages = {}
        for annotator_type, annotations in annotations_result.items():
            stages[annotator_type] = self._annotationFromJava(annotations)
        return stages


    def annotate(self, *args, **kwargs):
        """
        Annotate text(s) or text(s) with IDs using the LightPipeline.

        Supports both:
          - annotate(text: str)
          - annotate(texts: list[str])
          - annotate(ids: list[int], texts: list[str])
          - annotate(text: str, metadata: dict[str, list[str]])
          - annotate(texts: list[str], metadata: list[dict[str, list[str]]])
          - annotate(texts: list[str], metadata: dict[str, list[str]])

        Returns
        -------
        list[dict[str, list[str]]]

        Examples
        --------
        >>> from sparknlp.pretrained import PretrainedPipeline
        >>> explain_document_pipeline = PretrainedPipeline("explain_document_dl")
        >>> result = explain_document_pipeline.annotate('U.N. official Ekeus heads for Baghdad.')
        >>> result.keys()
        dict_keys(['entities', 'stem', 'checked', 'lemma', 'document', 'pos', 'token', 'ner', 'embeddings', 'sentence'])
        >>> result["ner"]
        ['B-ORG', 'O', 'O', 'B-PER', 'O', 'O', 'B-LOC', 'O']
        """
        metadata = kwargs.pop("metadata", None)

        def reformat(annotations):
            return {k: list(v) for k, v in annotations.items()}

        if "target" in kwargs:
            args = (kwargs["target"],) + args
        if "optional_target" in kwargs:
            args = args + (kwargs["optional_target"],)
        args, metadata = self.__extractMetadataFromArgs(args, metadata)

        stages = self.pipeline_model.stages
        if not self._skipPipelineValidation(stages):
            self._validateStagesInputCols(stages)

        input_type = self.__detectInputType(args)

        if input_type == "ids_texts":
            ids, texts = args
            if metadata is not None:
                raise TypeError("metadata is only supported for text inputs.")
            annotations = self._lightPipeline.annotateWithIdsJava(ids, texts)
            results = list(map(lambda a: reformat(a), list(annotations)))
            return results

        if input_type == "qa":
            question, context = args
            if metadata is not None:
                raise TypeError("metadata is only supported for text inputs.")
            if isinstance(question, list) and isinstance(context, list):
                annotations = self._lightPipeline.annotateJava(question, context)
                results = list(map(lambda a: reformat(a), list(annotations)))
                return results
            else:
                annotations = self._lightPipeline.annotateJava(question, context)
                results = reformat(annotations)
                return results

        if input_type == "text":
            target = args[0]
            metadata = self.__normalizeMetadataForTextInput(target, metadata)
            if isinstance(target, str):
                if metadata is None:
                    annotations = self._lightPipeline.annotateJava(target)
                else:
                    annotations = self._lightPipeline.annotateWithMetaJava(target, metadata)
                results = reformat(annotations)
                return results
            else:
                if metadata is None:
                    annotations = self._lightPipeline.annotateJava(target)
                    results = list(map(lambda a: reformat(a), list(annotations)))
                    return results
                if not isinstance(metadata, list):
                    raise TypeError(
                        "metadata must be a list of dicts or a dict of lists when "
                        "passing multiple texts."
                    )
                if len(metadata) != len(target):
                    raise TypeError("metadata and target must have the same length.")
                annotations = self._lightPipeline.annotateWithMetaJava(target, metadata)
                results = list(map(lambda a: reformat(a), list(annotations)))
                return results

        raise TypeError(
            "Unsupported input for annotate(). Expected: "
            "(text: str | list[str]), "
            "(ids: list[int], texts: list[str]), "
            "or (question: str, context: str)."
        )

    def __detectInputType(self, args):
        """
        Determine the input type pattern for fullAnnotate().
        Returns one of: 'ids_texts', 'qa', 'text', 'audio', 'image', or 'unknown'.
        """
        if len(args) == 2:
            a1, a2 = args

            # (ids, texts)
            if (
                    isinstance(a1, list)
                    and all(isinstance(i, int) for i in a1)
                    and isinstance(a2, list)
                    and all(isinstance(t, str) for t in a2)
            ):
                return "ids_texts"

            # (question, context)
            if isinstance(a1, str) and isinstance(a2, str):
                return "qa"

            # (questions[], contexts[])
            if (
                    isinstance(a1, list)
                    and all(isinstance(q, str) for q in a1)
                    and isinstance(a2, list)
                    and all(isinstance(c, str) for c in a2)
            ):
                return "qa"

        elif len(args) == 1:
            a1 = args[0]

            if not isinstance(a1, (str, list)):
                return "unknown"

            # 🧩 Case 1: plain string
            if isinstance(a1, str):
                if self.__isPath(a1):
                    return "image"
                if self.__isTextInput(a1):
                    return "text"
                return "unknown"

            # 🧩 Case 2: list — ensure homogeneous types
            if isinstance(a1, list) and len(a1) > 0:

                # Guard clause — mixed or invalid types
                if not all(isinstance(x, (str, float, list)) for x in a1):
                    return "unknown"

                # Text list
                if all(isinstance(x, str) for x in a1) and self.__isTextInput(a1):
                    return "text"

                # Audio list
                if all(isinstance(x, float) for x in a1) or (
                        all(isinstance(x, list) for x in a1) and all(isinstance(i, float) for sub in a1 for i in sub)
                ):
                    return "audio"

                # Image list (only strings allowed)
                if all(isinstance(x, str) for x in a1) and all("/" in x for x in a1):
                    return "image"

            return "unknown"

        return "unknown"


    def transform(self, dataframe):
        """Transforms a dataframe provided with the stages of the LightPipeline.

        Parameters
        ----------
        dataframe : :class:`pyspark.sql.DataFrame`
            The Dataframe to be transformed

        Returns
        -------
        :class:`pyspark.sql.DataFrame`
            The transformed DataFrame
        """
        transformed_df = self.pipeline_model.transform(dataframe)

        if self.output_cols:
            original_cols = dataframe.columns
            filtered_cols = list(dict.fromkeys(original_cols + self.output_cols))
            transformed_df = transformed_df.select(*filtered_cols)

        return transformed_df

    def setIgnoreUnsupported(self, value):
        """Sets whether to ignore unsupported AnnotatorModels.

        Parameters
        ----------
        value : bool
            Whether to ignore unsupported AnnotatorModels.

        Returns
        -------
        LightPipeline
            The current LightPipeline
        """
        self._lightPipeline.setIgnoreUnsupported(value)
        return self

    def getIgnoreUnsupported(self):
        """Gets whether to ignore unsupported AnnotatorModels.

        Returns
        -------
        bool
            Whether to ignore unsupported AnnotatorModels.
        """
        return self._lightPipeline.getIgnoreUnsupported()
